 /*
  * Copyright John E. Lloyd, 2003. All rights reserved. Permission
  * to use, copy, and modify, without fee, is granted for non-commercial 
  * and research purposes, provided that this copyright notice appears 
  * in all copies.
  *
  * This  software is distributed "as is", without any warranty, including 
  * any implied warranty of merchantability or fitness for a particular
  * use. The authors assume no responsibility for, and shall not be liable
  * for, any special, indirect, or consequential damages, or any damages
  * whatsoever, arising out of or in connection with the use of this
  * software.
  */

package convexhull3d;

import java.util.*;

/**
 * Test class for ConvexHull3D.
 *
 * @author John E. Lloyd, Winter 2003
 */
public class ConvexHull3DTest
{
	private ConvexHull3DTest()
	 { 
	 }

	private boolean facesEqual (int[] face1, int[] face2)
	 {
	   if (face1.length != face2.length)
	    { return false;
	    }
	   int len = face1.length;
	   int j;
	   for (j=0; j<len; j++)
	    { if (face1[0] == face2[j])
	       { break; 
	       }
	    }
	   if (j==len)
	    { return false;
	    } 
	   for (int i=1; i<len; i++)
	    { if (face1[i] != face2[(j+i)%len])
	       { return false; 
	       }
	    }
	   return true;	   
	 }

	private void singletest (double[] coords, int[][] checkFaces)
	   throws Exception
	 {
	   ConvexHull3D hull = new ConvexHull3D (coords);
	   int [][] mergedFaces = hull.getMergedFaces();
	   if (mergedFaces.length != checkFaces.length)
	    { throw new Exception (
"Error: " + mergedFaces.length + " merged faces vs. " + checkFaces.length);
	    }
	   // translate merged face indices back into original indices
	   SpatialPoint[] vtxs = hull.getMergedFaceVertices();
	   for (int j=0; j<mergedFaces.length; j++)
	    { int[] mf = mergedFaces[j];
	      for (int k=0; k<mf.length; k++)
	       { mf[k] = vtxs[mf[k]].getPointIndex(); 
	       }
	    }
	   for (int i=0; i<checkFaces.length; i++)
	    { int[] cf = checkFaces[i];
	      int j;
	      for (j=0; j<mergedFaces.length; j++)
	       { if (mergedFaces[j] != null)
		  { if (facesEqual (cf, mergedFaces[j]))
		     { mergedFaces[j] = null;
		       break;
		     }
		  }
	       }
	      if (j == mergedFaces.length)
	       { String s = "";
		 for (int k=0; k<cf.length; k++)
		  { s += cf[k] + " ";
		  }
		 throw new Exception ("Error: face " + s + " not found");
	       }
	    }
	 }

	private void rotateCoords (double[] res, double[] xyz,
			      double roll, double pitch, double yaw)
	 {
	   double sroll = Math.sin (roll);
	   double croll = Math.cos (roll);
	   double spitch = Math.sin (pitch);
	   double cpitch = Math.cos (pitch);
	   double syaw = Math.sin (yaw);
	   double cyaw = Math.cos (yaw);

	   double m00 = croll * cpitch;
	   double m10 = sroll * cpitch;
	   double m20 = - spitch;

	   double m01 = croll * spitch * syaw - sroll * cyaw;
	   double m11 = sroll * spitch * syaw + croll * cyaw;
	   double m21 = cpitch * syaw;

	   double m02 = croll * spitch * cyaw + sroll * syaw;
	   double m12 = sroll * spitch * cyaw - croll * syaw;
	   double m22 = cpitch * cyaw;

	   double x, y, z;

	   for (int i=0; i<xyz.length-2; i+=3)
	    {
	      res[i+0] = m00*xyz[i+0] + m01*xyz[i+1] + m02*xyz[i+2];
	      res[i+1] = m10*xyz[i+0] + m11*xyz[i+1] + m12*xyz[i+2];
	      res[i+2] = m20*xyz[i+0] + m21*xyz[i+1] + m22*xyz[i+2];
	    }
	 }

	private void test (double[] coords, int[][] checkFaces)
	   throws Exception
	 { 
	   double[][] rpyList = new double[][]
	    { 
	      {  0,  0,  0},
	      { 10, 20, 30},
	      { -45, 60, 91},
	      { 125, 67, 81}
	    };
	   double[] xcoords = new double[coords.length];

	   for (int i=0; i<rpyList.length; i++)
	    { double[] rpy = rpyList[i]; 
	      rotateCoords (xcoords, coords,
			    Math.toRadians(rpy[0]),
			    Math.toRadians(rpy[1]),
			    Math.toRadians(rpy[2]));
	      singletest (coords, checkFaces);
	    }
	 }

	private void setTriangleMarks (Triangle[] tris, int[] marks)
	 {
	   for (int i=0; i<tris.length; i++)
	    { tris[i].setMark (marks[i]);
	    }
	 }

	private String markString (int[] marks)
	 {
	   String s = "";
	   for (int i=0; i<marks.length; i++)
	    { s += marks[i];
	      if (i<marks.length-1)
	       { s += " "; 
	       }
	    }
	   return s;
	 }

	private void checkTriangleMarks (Triangle[] tris, int[] marks)
	   throws Exception
	 {
	   int[] trimarks = new int[tris.length];
	   for (int i=0; i<tris.length; i++)
	    { trimarks[i] = tris[i].getMark();
	    }
	   int i;
	   for (i=0; i<trimarks.length; i++)
	    { if (trimarks[i] != marks[i])
	       { break;
	       }
	    }
	   if (i<trimarks.length)
	    { throw new Exception (
"got triangle marks " + markString(trimarks) + " vs. " + markString(marks));
	    }
	 }

	private void crossedFaceTest ()
	   throws Exception
	 { 
	   // create a "test" triangle set
	   SpatialPoint vtxs[] = new SpatialPoint[5];
	   vtxs[0] = new SpatialPoint (0, 0, 0, 0);
	   vtxs[1] = new SpatialPoint (1, -1, 0, 1);
	   vtxs[2] = new SpatialPoint (0, 1, 0, 2);
	   vtxs[3] = new SpatialPoint (-1, -1, 0, 3);
	   vtxs[4] = new SpatialPoint (0, 0, -1, 4);

	   Triangle tris[] = new Triangle[6];

	   tris[0] = new Triangle (vtxs[0], vtxs[3], vtxs[1]);
	   tris[1] = new Triangle (vtxs[0], vtxs[1], vtxs[2]);
	   tris[2] = new Triangle (vtxs[0], vtxs[2], vtxs[3]);

	   tris[3] = new Triangle (vtxs[4], vtxs[1], vtxs[3]);
	   tris[4] = new Triangle (vtxs[4], vtxs[2], vtxs[1]);
	   tris[5] = new Triangle (vtxs[4], vtxs[3], vtxs[2]);

	   // connect the edges by brute force
	   for (int i=0; i<tris.length; i++)
	    { for (int k=0; k<3; k++)
	       { HalfEdge e = tris[i].getEdge(k);
		 if (e.getOpposite() == null)
		  { HalfEdge o = null;
		    for (int j=0; j<tris.length; j++)
		     { if ((o = tris[j].findEdge (e.head(), e.tail())) != null)
			{ e.setOpposite (o);
			  o.setOpposite (e);
			  break;
			}
		     }
		  }
	       }
	    }

	   int U = Triangle.COPLANAR;
	   int V = Triangle.VISIBLE;
	   int N = Triangle.NO_MARK;

	   HalfEdge edge0 = tris[3].findEdge (vtxs[4], vtxs[1]);
	   ConvexHull3D hull = new ConvexHull3D (vtxs, tris);

	   setTriangleMarks (tris, new int[] { U, U, N, V, N, N });
	   hull.unmarkImproperFaces (edge0, new SpatialPoint (0, -2, 0, -1));
	   checkTriangleMarks (tris, new int[] { U, N, N, V, N, N });

	   setTriangleMarks (tris, new int[] { U, U, N, V, N, N });
	   hull.unmarkImproperFaces (edge0, new SpatialPoint (1, -2, 0, -1));
	   checkTriangleMarks (tris, new int[] { U, U, N, V, N, N });

	   setTriangleMarks (tris, new int[] { U, U, U, V, N, N });
	   hull.unmarkImproperFaces (edge0, new SpatialPoint (0, -2, 0, -1));
	   checkTriangleMarks (tris, new int[] { U, U, U, V, N, N });
	 }

	private void attachTriangles (Triangle tri1, Triangle tri2,
				      SpatialPoint v1, SpatialPoint v2)
	 {
	   HalfEdge e = tri1.findEdge (v1, v2);
	   HalfEdge o = tri2.findEdge (v2, v1);
	   e.setOpposite (o);
	   o.setOpposite (e);
	 }

	private void mergedFaceTest  ()
	   throws Exception
	 { 
	   // create a "test" triangle set
	   SpatialPoint vtxs[] = new SpatialPoint[10];
	   vtxs[0] = new SpatialPoint (0, 0, 0, 0);
	   vtxs[1] = new SpatialPoint (1, 0, 0, 1);
	   vtxs[2] = new SpatialPoint (0.5, 1, 0, 2);
	   vtxs[3] = new SpatialPoint (-0.5, 1, 0, 3);
	   vtxs[4] = new SpatialPoint (-1, 0, 0, 4);
	   vtxs[5] = new SpatialPoint (-0.5, -1, 0, 5);
	   vtxs[6] = new SpatialPoint (0.5, -1, 0, 6);
	   vtxs[7] = new SpatialPoint (1.5, 1, 0, 7);
	   vtxs[8] = new SpatialPoint (2, 0, 0, 8);
	   vtxs[9] = new SpatialPoint (1.5, -1, 0, 9);

	   Triangle tris[] = new Triangle[10];
	   for (int i=0; i<6; i++)
	    { tris[i] = new Triangle (vtxs[0], vtxs[i+1], vtxs[((i+1)%6)+1]);
	    }
	   tris[6] = new Triangle (vtxs[1], vtxs[7], vtxs[2]);
	   tris[7] = new Triangle (vtxs[1], vtxs[8], vtxs[7]);
	   tris[8] = new Triangle (vtxs[1], vtxs[9], vtxs[8]);
	   tris[9] = new Triangle (vtxs[1], vtxs[6], vtxs[9]);

	   for (int i=0; i<6; i++)
	    { attachTriangles (tris[i], tris[(i+5)%6], vtxs[0], vtxs[i+1]);
	    }
	   attachTriangles (tris[0], tris[6], vtxs[1], vtxs[2]);
	   attachTriangles (tris[6], tris[7], vtxs[1], vtxs[7]);
	   attachTriangles (tris[7], tris[8], vtxs[1], vtxs[8]);
	   attachTriangles (tris[8], tris[9], vtxs[1], vtxs[9]);
	   attachTriangles (tris[9], tris[5], vtxs[1], vtxs[6]);

	   ConvexHull3D hull = new ConvexHull3D (vtxs, tris);

	   HalfEdge edge0 = tris[4].findEdge (vtxs[5], vtxs[6]);
	   SpatialVector nrml = new SpatialVector (0, 0, 1);

	   setTriangleMarks (tris, new int[] { 3,0,3,3,3,3,3,0,0,0 });
	   edge0 = hull.unmarkNonconvexFaces (edge0, nrml, 3);
	   checkTriangleMarks (tris, new int[] { 3,0,2,2,3,3,3,0,0,0 });
	   hull.remarkRegion (edge0.getFace(), 3, 1);
	   checkTriangleMarks (tris, new int[] { 1,0,2,2,1,1,1,0,0,0 });
	   hull.remarkNonconvexRegion (edge0, 1);
	   checkTriangleMarks (tris, new int[] { 1,0,0,0,1,1,1,0,0,0 });

	   edge0 = tris[9].findEdge (vtxs[6], vtxs[9]);
	   setTriangleMarks (tris, new int[] { 3,3,3,0,0,3,0,3,3,3 });
	   hull.unmarkNonconvexFaces (edge0, nrml, 3);
	   checkTriangleMarks (tris, new int[] { 2,3,3,0,0,2,0,3,3,3 });
	   hull.remarkRegion (edge0.getFace(), 3, 1);
  	   checkTriangleMarks (tris, new int[] { 2,3,3,0,0,2,0,1,1,1 });
	   hull.remarkNonconvexRegion (edge0, 1);
  	   checkTriangleMarks (tris, new int[] { 0,0,0,0,0,0,0,1,1,1 });

  	   edge0 = tris[5].findEdge (vtxs[0], vtxs[6]);
  	   setTriangleMarks (tris, new int[] { 3,3,3,0,0,3,0,3,3,3 });
  	   edge0 = hull.unmarkNonconvexFaces (edge0, nrml, 3);
  	   checkTriangleMarks (tris, new int[] { 2,3,3,0,0,2,0,3,3,3 });

  	   edge0 = tris[7].findEdge (vtxs[7], vtxs[1]);
  	   setTriangleMarks (tris, new int[] { 3,3,3,0,3,3,0,3,3,3 });
  	   edge0 = hull.unmarkNonconvexFaces (edge0, nrml, 3);
  	   checkTriangleMarks (tris, new int[] { 2,3,3,0,3,2,0,3,3,3 });
	   hull.remarkRegion (edge0.getFace(), 3, 1);
  	   checkTriangleMarks (tris, new int[] { 2,3,3,0,3,2,0,1,1,1 });
	   hull.remarkNonconvexRegion (edge0, 1);
  	   checkTriangleMarks (tris, new int[] { 0,0,0,0,0,0,0,1,1,1 });

	 }

	/**
	 * Runs a set of tests on the ConvexHull3D class, and
	 * prints <code>Passed</code> if all is well.
	 * Otherwise, an error message and stack trace
	 * are printed.
	 */
	public static void main (String[] args) 
	 {
	   ConvexHull3DTest tester = new ConvexHull3DTest();

	   try
	    { 
	      tester.test (
		 new double[] {
		   0,    0,    0,
		   0.9,  0.9,  0,
		   -0.9, -0.8, 0,
		   1,    1,    1,
		   1,   -1,    1,
		   0,   -1,    1,
		   1,    1,   -1,
		   1,   -1,   -1,
		   1,    0,   -1,
		   -1,    1,    1,
		   -1,   -1,    1,
		   -1,   -1,    0,
		   -1,    1,   -1,
		   -1,   -1,   -1 
		 },
		 new int[][] {
		   { 10, 4, 3, 9 },
		   { 12, 9, 3, 6 }, 
		   { 13, 10, 9, 12 }, 
		   { 6, 3, 4, 7 }, 
		   { 7,  4, 10, 13},
		   { 13, 12, 6, 7 },
		 });

	      double[] pnts = new double[] {
-1.2232872033130677, 1.1187955883870726, -0.5950343714561437,
1.2232872033130677, -1.1187955883870726, 0.5950343714561437,
0.357261799528629, -1.6187955883870726, 0.5950343714561437,
0.28359458252715936, 0.508799774311675, 0.0026381060040960302,
-0.28359458252715936,-0.508799774311675,-0.0026381060040960302,
-0.7534408929201135, 0.3049979070376989, -0.9370545147818123,
0.7534408929201135, -0.3049979070376989, 0.9370545147818123,
-1.0522771316502335, 0.8225974556610488, 0.3446582493297647,
0.18625172786579475, -1.3225974556610487, -0.3446582493297647,
0.5824308212572793,-0.008799774311675002,-0.0026381060040960302,
-0.5824308212572793,0.008799774311675002,0.0026381060040960302,
-0.18625172786579475, 1.3225974556610487, 0.3446582493297647,
1.0522771316502335, -0.8225974556610488, -0.3446582493297647,
-0.357261799528629, 1.6187955883870726, -0.5950343714561437,
-0.11258451086432514, -0.8049979070376988, 0.9370545147818123,
0.11258451086432514, 0.8049979070376988, -0.9370545147818123
	      };

	      tester.test (pnts,
		 new int[][] {
		   { 15, 5, 0, 13},
		   { 8, 5, 15, 12},
		   { 8, 12, 1, 2},
		   { 13, 0, 7, 11},
		   { 12, 15, 13, 11, 6, 1},
		   { 5, 8, 2, 14, 7, 0},
		   { 14, 2, 1, 6},
		   { 6, 11, 7, 14},
		 });

	      pnts = new double[] {
0.9714990580778444, -0.02248791591143462, -0.004642869064161337, 
0.6413474015640961, -0.9871842324649125, 0.022044414088051578, 
-0.7231847466470569, 0.5123482392443048, -0.22748957518627289, 
0.3765860431492156, 0.7534157914560631, 0.7150110332984456, 
-0.925143186534273, -0.5648918210590976, -0.23491954889192335, 
0.7729239467329101, 0.10116871116131732, -0.6975466973230646, 
-0.9583759086133905, -0.3382918570249631, -0.6941439237990923, 
-0.5408497858189025, -0.19024717076241893, -0.666242228568849, 
0.2242679002160075, 0.03745214796957619, 0.8099455779733638, 
0.08250585015063283, -0.38474620180115493, -0.8772781061369301, 
0.862439972884258, 0.27068968683565253, -0.335420812676283, 
0.23061506063957604, -0.06400192236218394, -0.035329416583015805, 
-0.17896653575877175, 0.9093418201993233, 0.5500775676801679, 
-0.9030621259485898, 0.24527120459361118, -0.7223854798874207, 
-0.21950919845094163, -0.07849235987363223, -0.17786735503927886, 
0.23271756697964396, 0.14058396641935467, 0.7320032350207677, 
-0.7796102950222092, -0.950536376109878, -0.8255758348707258, 
-0.9202163611960321, -0.4204302746848856, -0.8603716602258262, 
0.7254336165782997, 0.059563102281323665, 0.06170970010098897, 
0.8527841491617174, -0.4666353306620987, 0.6225446600584299, 
	      };

	      tester.test (
		 pnts,
	         new int[][] {
		   { 5, 0, 1},
		   { 13, 12, 5},
		   { 13, 6, 4},
		   { 4, 6, 16},
		   { 16, 1, 4},
		   { 5, 12, 10},
		   { 10, 12, 3},
		   { 10, 0, 5},
		   { 3, 0, 10},
		   { 8, 12, 4},
		   { 3, 12, 8},
		   { 4, 12, 2},
		   { 2, 12, 13},
		   { 13, 4, 2},
		   { 5, 1, 9},
		   { 9, 1, 16},
		   { 13, 5, 9},
		   { 4, 1, 19},
		   { 8, 4, 19},
		   { 1, 0, 19},
		   { 19, 0, 3},
		   { 19, 3, 8},
		   { 17, 13, 9},
		   { 9, 16, 17},
		   { 17, 6, 13},
		   { 16, 6, 17},
		 });		   

	      pnts = new double[] {
0.001605132494626904, 0.030721751610423897, -0.8411476171776633, 
0.5634253525903541, 0.8443226461534035, 0.967938854586033, 
0.6618125512293167, -0.275360478426135, -0.007468490341031586, 
-0.176696205588317, -0.8505163095954686, 0.031894012140315287, 
0.8769237651669879, 0.22274602263265253, 0.7618547588785909, 
0.7295950803728364, 0.7667271697469178, 0.14263093021739026, 
0.04681987119537512, 0.2190740622071501, -0.42986781505116545, 
0.9221317718793494, 0.8061267707764119, 0.38638593741559557, 
0.29273299720521595, 0.8837041038906073, -0.4647168784291016, 
-0.17208554857438862, 0.14940530442774724, 0.3622616321879668, 
-0.9485236734465277, 0.1308262660907511, 0.8504335923965329, 
0.1096662654636189, -0.664798536576406, 0.687684338487782, 
0.42059189520755846, -0.9803373495155727, -0.6445647036670425, 
0.6592943653314729, -0.9397281248674123, -0.09059832043767968, 
-0.2951677768548895, -0.3528748995424238, 0.02064341918376189, 
-0.3027863518881351, -0.7594398040262234, -0.6035914421728679, 
-0.7597770881645689, 0.7323078472196445, -0.39196467309724814, 
-0.5767089274299881, 0.5781156177495206, -0.7394103860258159, 
-0.48104658151147883, -0.15582670810165, 0.8795887511875711, 
-0.9481968218495915, -0.03834159245063273, 0.8826412504794428, 
0.8225225188048824, -0.933277012162482, 0.14102780654979052, 
0.5391572346901057, -0.4813227771272295, -0.29086775229164696, 
-0.1233040654269566, -0.26905547880292824, -0.818394686743154, 
0.6858436121310094, -0.5674235168066306, -0.15573407335149092, 
0.8234269164018719, 0.78903131466176, 0.3072859270481083, 
-0.9114392122076171, 0.9055793648112069, 0.018660777166646048, 
0.3842113851957629, -0.698327859058445, 0.14404101150341564, 
-0.6211829484044935, 0.02160537320909528, -0.0723396098139577, 
-0.7425796406688165, -0.5897512344094613, 0.15723195429857406, 
0.773465400202374, 0.34669228138153474, -0.9375397448948906, 
	      };

	      tester.test (
		 pnts,
		 new int[][] {
		   { 29, 7, 20},
		   { 20, 12, 29},
		   { 8, 7, 29},
		   { 25, 10, 1},
		   { 1, 7, 8},
		   { 8, 25, 1},
		   { 29, 12, 22},
		   { 22, 12, 15},
		   { 20, 7, 4},
		   { 4, 7, 1},
		   { 19, 10, 25},
		   { 19, 25, 28},
		   { 1, 10, 19},
		   { 28, 25, 16},
		   { 16, 25, 8},
		   { 15, 12, 3},
		   { 28, 15, 3},
		   { 3, 12, 20},
		   { 22, 15, 17},
		   { 17, 15, 28},
		   { 17, 28, 16},
		   { 8, 29, 17},
		   { 16, 8, 17},
		   { 18, 1, 19},
		   { 19, 28, 11},
		   { 11, 28, 3},
		   { 18, 19, 11},
		   { 3, 20, 11},
		   { 11, 20, 4},
		   { 4, 1, 11},
		   { 11, 1, 18},
		   { 0, 29, 22},
		   { 17, 29, 0},
		   { 0, 22, 17},
		  });

	      tester.test (
		 new double[] {
		   0,    0,    0,
		   0.9,  0.9,  0,
		   -0.9, -0.8, 0,
		   0.5,  0.5,  1,
		   -0.5, 0.5,  1, 
		   0.5, -0.5,  1,
		   -0.5, -0.5,  1, 
		   0,    0,    1,
		   1,    1,    1,
		   1,   -1,    1,
		   0,   -1,    1,
		   -1,    1,    1,
		   -1,   -1,    1,
		   1,   -0.5,  0.5,
		   1,    0.5,  0.5,
		   1,   -0.5, -0.5,
		   1,    0.5, -0.5,
		  -1,   -0.5,  0.5,
		  -1,    0.5,  0.5,
		  -1,   -0.5, -0.5,
		  -1,    0.5, -0.5,
		   0,    0,    -1,
		   0.5,  0.5,  -1,
		   -0.5, 0.5,  -1, 
		   0.5, -0.5,  -1,
		   -0.5, -0.5, -1, 
		   1,    1,   -1,
		   1,   -1,   -1,
		   -1,    1,   -1,
		   -1,   -1,   -1 
		 },
		 new int[][] {
		   { 12, 9, 8, 11},
		   { 26, 8, 9, 27},
		   { 27, 9, 12, 29},
		   { 26, 27, 29, 28},
		   { 29, 12, 11, 28},
		   { 28, 11, 8, 26},
		 });

	      tester.crossedFaceTest();
	      tester.mergedFaceTest();
		   
	    }
	   catch (Exception e) 
	    { e.printStackTrace();
	      System.exit(1); 
	    }

	   System.out.println ("\nPassed\n");
	 }
}
